#!/usr/bin/env python3

import logging
import multiprocessing as mp
import sys
import psutil
import re
import resource
import subprocess
import time


def check_free_command():
    err = subprocess.run(['free', '--mega'], capture_output=True, shell=True).stderr

    if err:
        return False
    else:
        return True


def limit(mem_limit):
    """
    Defines a wrapper for functions that first sets a resource limit before
    executing the function.

    :param mem_limit:
    :return:
    """
    def inner(fn):
        """
        Wrapper that sets some memory limit then executes the input function.

        :param fn:
        :return:
        """
        def limited_fn(*args, **kwargs):
            soft, hard = resource.getrlimit(resource.RLIMIT_RSS)
            resource.setrlimit(resource.RLIMIT_RSS, (mem_limit, hard))

            fn(*args, **kwargs)

        return limited_fn
    return inner


def sandbox(logging_name, mem_limit, interval):
    """
    Create a decorator that runs function in a new process, limiting the resources of that
    process and observing it at regular intervals.

    :param mem_limit:
    :param interval:
    :return:
    """
    def inner(fn):
        """
        Run function in a new process, modifying the function so that it first sets a resource
        limit for the environment it is in.

        :param fn:
        :return:
        """
        def job(*args, **kwargs):
            # Configure logging output.
            logger = logging.getLogger(logging_name)
            logger.info("\ttotal\tused\tfree\tshared\tbuff/cache\tavailable")

            # Initial log for starting memory usage.
            mem_info = subprocess.run(['free', '--mega'], capture_output=True).stdout.decode()
            logger.info("Initial...%s" % "\t".join(re.findall(r"[^\s]+", mem_info)[6:13]))

            # Start new child process.
            ctx = mp.get_context('fork')

            new_target = limit(mem_limit)(fn)
            p = ctx.Process(target=new_target, args=args, kwargs=kwargs)

            p.start()

            while p.is_alive():
                # Log memory usage.
                mem_info = subprocess.run(['free', '--mega'], capture_output=True).stdout.decode()
                logger.info("\t".join(re.findall(r"[^\s]+", mem_info)[6:13]))

                time.sleep(interval)

            p.join()

        return job
    return inner


def parse():
    from argparse import ArgumentParser, RawTextHelpFormatter, REMAINDER

    parser = ArgumentParser(description="::: expam_limit ::: \n\n\tcommand for limiting and logging memory usage.", formatter_class=RawTextHelpFormatter)
    parser.add_argument('-m', '--memory', type=int, dest='memory', help="Explicit memory limit in bytes.")
    parser.add_argument('-x', '--x', type=float, dest='x', help="Memory limit as a percentage of total system memory.")
    parser.add_argument('-t', '--interval', type=float, dest='interval', default=1.0, help="Time between memory logs.")
    parser.add_argument('-o', '--out', type=str, dest='out', help="File to log to (default is stdout).")
    parser.add_argument('command', nargs=REMAINDER, help="Command to run in limited regime.")

    args = parser.parse_args()

    # Establish memory limit.
    byte_limit = args.memory
    perc_limit = None if args.x is None else int(args.x * psutil.virtual_memory().total)

    if byte_limit is not None and perc_limit is not None:
        mem_limit = min(byte_limit, perc_limit)

    elif byte_limit is not None:
        mem_limit = byte_limit
    
    elif perc_limit is not None:
        mem_limit = byte_limit
    
    else:
        mem_limit = None

    return mem_limit, args.interval, args.out, args.command


def sandbox_command():
    mem_limit, interval, out_file, cmd = parse()

    if mem_limit is None:
        soft, hard = resource.getrlimit(resource.RLIMIT_RSS)
        mem_limit = soft

    logger = logging.getLogger('sandbox_application')
    logger.setLevel(logging.DEBUG)

    # Create the appropriate handler.
    if out_file is None:
        handle = logging.StreamHandler()
    else:
        handle = logging.FileHandler(out_file)

    handle.setLevel(logging.DEBUG)
    formatter = logging.Formatter('%(asctime)s ... %(message)s')
    handle.setFormatter(formatter)

    logger.addHandler(handle)

    @sandbox(logging_name='sandbox_application', mem_limit=mem_limit, interval=interval)
    def run_cmd(command):
        subprocess.run(command, shell=False, check=True)

    if cmd:
        run_cmd(cmd)


def main():
    if not check_free_command():
        print("Must run on Linux system (or have `free` command available).")
        sys.exit(1)
    
    sandbox_command()


if __name__ == '__main__':
    main()



import io
import re
from expam.database import TIMEOUT
from expam.ext.kmers import bytes_intersect
from expam.process import COMMAND_CODES
from expam.process.jobworker import JobWorker
from expam.sequences import format_name, get_opener


class ReadWorker(JobWorker):
    def __init__(
        self, paired_mode, n_classifiers, logging_dir, status_index, statuses, child_statuses,
        timeout=TIMEOUT, phases=None
    ):
        super(ReadWorker, self).__init__(
            logging_dir, status_index, statuses, child_statuses,
            timeout=timeout, phases=phases
        )
        self.job_name = "reader"

        self.paired_mode = paired_mode
        self.n_classifiers = n_classifiers
        self._c = 0  # Which classifier do we next send data to?

        self.n_streams = 1 if not paired_mode else 2

        self.init_methods_from_dict(
            {
                "classify": {
                    "getter": None,
                    "processor": self.load_reads,
                    "transition": None,
                }
            }
        )

    def load_reads(self, file_urls):
        print("Loading reads from %s..." % ", ".join(file_urls))

        file_names, file_modes, file_openers, file_types = self.init_openers(*file_urls)

        # Open each file.
        _file_handles = [
            opener(file_url, mode)
            for opener, file_url, mode in zip(file_openers, file_urls, file_modes)
        ]

        read_stream = io.BytesIO()

        read_ids = []
        read_seq = None

        while read_ids or read_seq is None:  # read_seq is None is initial case.
            for i in range(self.n_streams):
                file_type = file_types[i]

                if file_type == "fasta":
                    next_read_id = self.scan_fa_to_next_read(_file_handles[i], read_stream)

                else:  # File type is "fastq".
                    next_read_id = self.scan_fq_to_next_read(_file_handles[i], read_stream)

                if next_read_id is not None:
                    if i == 0:
                        read_ids.append(next_read_id)
                    else:
                        read_ids[-1] = bytes_intersect(read_ids[-1], next_read_id)

            read_seq = read_stream.getvalue()

            if len(read_seq) > self.n_streams:
                read_id = read_ids.pop(0)

                self.send_read(min(file_names), read_id, read_seq)

            read_stream.seek(0)
            read_stream.truncate()

        for i in range(self.n_streams):
            # Close file handles.
            _file_handles[i].close()

        # Close data stream.
        del read_stream

    def send_read(self, file_name, read_id, read):
        self.tell_child(
            cid=self._c,
            data=(
                COMMAND_CODES["ACTIVE"],
                (file_name, read_id, read, self.child_id)
            )
        )

        self._c = (self._c + 1) % self.n_classifiers

    def init_openers(self, *file_urls):
        file_names, file_modes, file_openers, file_types = [], [], [], []

        for file_url in file_urls:
            file_name = format_name(file_url)
            file_name, (mode, opener) = get_opener(file_name)

            if self.check_suffix(file_url, ".fasta", ".fa"):
                file_type = "fasta"

            elif self.check_suffix(file_url, ".fastq", ".fq"):
                file_type = "fastq"

            else:
                raise ValueError("Unknown sequence format for %s!" % file_url)

            file_names.append(file_name)
            file_modes.append(mode)
            file_openers.append(opener)
            file_types.append(file_type)

        return file_names, file_modes, file_openers, file_types

    @staticmethod
    def check_suffix(string, *suffixes):
        for suffix in suffixes:
            if re.findall(r"(%s)(?:.tar)*(?:.gz)*$" % suffix, string):
                return True

        return False

    @staticmethod
    def scan_fa_to_next_read(file, read_stream):
        next_read_id = None

        while True:
            line = file.readline()

            if line == b'':
                read_stream.write(b'N')
                break  # EOF.

            if b'>' in line:
                next_read_id = re.match(b'^>(\S+)\s', line).group(1)

                # End my contribution to this read.
                read_stream.write(b'N')
                break

            else:
                read_stream.write(line.strip())

        return next_read_id

    @staticmethod
    def scan_fq_to_next_read(file, read_stream):
        next_read_id = None

        while True:
            line = file.readline()

            if line == b'':
                read_stream.write(b'N')
                break

            if b'@' in line:
                next_read_id = re.match(b'^@(.*)\s', line).group(1)

                read_stream.write(b'N')
                break

            else:
                read_stream.write(line.strip())

                # Skip next two lines.
                for _ in range(2):
                    file.readline()

        return next_read_id




from ._build.kmers import \
    reverse_complement, \
    get_raw_kmers, \
    bytes_intersect, \
    reverse_complement_combine, \
    init_index, \
    remove_duplicates, \
    to_32bit, \
    import_mask, \
    map_kmers, \
    extract_kmers_from_read



import multiprocessing
import platform
from typing import Tuple

from expam.cli.build import BuildCommand
from expam.cli.classify import ClassifyCommand
from expam.cli.main import CommandGroup, ExpamOptions, retrieve_arguments
from expam.cli.tree import TreeCommand
from expam.cli.utils import UtilsCommand
from expam.utils import die


def main():
    # Set process creation method.
    if platform.system() != "Windows":
        multiprocessing.set_start_method("fork")

    args: ExpamOptions = retrieve_arguments()

    handlers: Tuple[CommandGroup] = (BuildCommand, ClassifyCommand, TreeCommand, UtilsCommand)
    for handler in handlers:
        if args.command in handler.commands:
            handler(**handler.take_args(args)).run(args.command)
            break
    else:
        die("Unrecognised command: %s" % args.command)


if __name__ == "__main__":
    main()
        


import os
import sys

from expam.sequences import COMPRESSION_EXTNS

def die(msg):
    print("ERROR\n", msg, "\n")
    sys.exit(1)

def yield_csv(path):
    with open(path, 'r') as f:
        for line in f:
            line = line.strip()
            if line:
                line = line.split(",")
                yield line

def ls(path, ext=None):
    # Get list of files.
    if os.path.isfile(path):
        files = [path]
    else:
        files = [
            os.path.join(path, file)
            for file in os.listdir(path)
            if not file.startswith(".")
        ]
        
        # Filter for files not subdirectories.
        files = [file for file in files if os.path.isfile(file)]

    if ext is not None:
        # Reduce list to those that have the right file extension.
        if not isinstance(ext, str) or isinstance(ext, list):
            raise TypeError("Argument `ext` must be a string of list of strings!")

        fcheck = isfiletype if isinstance(ext, str) else isfiletypes
        files = [file for file in files if fcheck(file, ext)]

    return files

def isfiletype(path: str, ext: str, check_compressed=False):
    if check_compressed:
        path = removecompressedextension(path)

    if path.endswith(ext):
        return True
    else:
        return False

def isfiletypes(path, ext, check_compressed=False):
    if check_compressed:
        path = removecompressedextension(path)

    return any(isfiletype(path, e) for e in ext)

def removecompressedextension(path):
    for cmp_ext in COMPRESSION_EXTNS:
        path = path.rstrip(cmp_ext)

    return path

def make_path_absolute(dir, base_dir):
    if base_dir not in dir:
        dir = os.path.join(base_dir, dir)

    return dir

def is_hex(string):
    if string[0] != "#" or len(string) != 7:
        return False

    try:
        int(string[1:].upper(), 16)
    except TypeError:
        return False

    return True

def parse_int(param):
    INVALID_PARAM_MSG = ("Invalid parameter (%s), must be integer!" % str(param))
    if param is not None:
        try:
            # Convert to float.
            param = float(param)
        except ValueError:
            die(INVALID_PARAM_MSG)

        # Convert to int and see if the value changes.
        new_param = int(param)
        if new_param != param:
            die(INVALID_PARAM_MSG)
    else:
        new_param = None

    return new_param

def parse_float(param):
    INVALID_PARAM_MSG = ("Invalid parameter (%s), must be integer!" % str(param))
    if param is not None:
        try:
            param = float(param)
        except ValueError:
            die(INVALID_PARAM_MSG)

    return param




from expam.process.manager import ControlCenter


class ExpamClassifierProcesses(ControlCenter):
    @classmethod
    def from_method_dict(cls, logging_dir, config):
        base_center = super(ExpamClassifierProcesses, cls).from_dict(logging_dir, config)

        base_arguments = {
            "group_name": "group_name",
            "workers": "workers",
            "child_statuses": "_child_statuses",
            "phases": "phases",
            "phase_queues": "phase_queues",
            "child_queues": "child_queues",
            "children": "_children",
            "processors": "_processors",
            "transitions": "_transitions",
            "timeout": "_timeout",
        }
        base_attributes = {
            attr: getattr(base_center, attr_reference)
            for attr, attr_reference in base_arguments.items()
        }

        # Create child class from this instance.
        control_center = ExpamClassifierProcesses(logging_dir=logging_dir, **base_attributes)

        control_center.set_methods(
            {
                "classify": {
                    "processor": None,
                    "transition": None
                }
            }
        )

        return control_center




from multiprocessing import shared_memory

import numpy as np
import tables as tb

from expam.database import DataTypeConfig, FileLocationConfig, expam_dtypes
from expam.database.config import load_database_config


class TablesDb:
    def __init__(self, db_url, shape=None, keys_type=None, values_type=None, create=False, edit=False):
        self.url = db_url

        self.keys_shape = shape
        self.values_shape = None if shape is None else tuple((shape[0],))

        self.keys_type = keys_type
        self.values_type = values_type

        self.edit = True if (create or edit) else False

        self.keys_offset = 0
        self.values_offset = 0

        filters = tb.Filters(complevel=5, complib="zlib")

        if create:
            if any([shape is None, keys_type is None, values_type is None]):
                raise ValueError("Parameters not provided for creation!")

            self.fh: tb.File = tb.open_file(db_url, mode="w")

            self.keys = self.fh.create_carray(self.fh.root, "keys", tb.UInt64Atom(),
                                              self.keys_shape, filters=filters)
            self.values = self.fh.create_carray(self.fh.root, "values", tb.UInt16Atom(),
                                                self.values_shape, filters=filters)

        else:
            if edit:
                self.fh: tb.File = tb.open_file(db_url, mode="r+")

                self.keys = self.fh.root.keys
                self.values = self.fh.root.values

            else:
                self.fh: tb.File = tb.open_file(db_url, mode="r")

    def _write(self, src, dest, offset):
        if not self.edit:
            raise IOError("Opened in read mode!")

        src_length = src.shape[0]
        dest[offset:offset + src_length] = src[:]

        return src_length

    def write_keys(self, arr, offset=0):
        return self._write(src=arr, dest=self.keys, offset=offset)

    def write_values(self, arr, offset=0):
        return self._write(src=arr, dest=self.values, offset=offset)

    def read_keys(self, arr):
        if self.edit:
            raise IOError("Opened in create mode!")

        self.fh.root.keys.read(out=arr)

    def read_values(self, arr):
        if self.edit:
            raise IOError("Opened in create mode!")

        self.fh.root.values.read(out=arr)

    def close(self):
        self.fh.close()


class SharedDB:
    def __init__(self, out_dir, keys_shape, values_shape, dtypes: DataTypeConfig = expam_dtypes, config: FileLocationConfig = None):
        print("Preparing shared memory allocations...")

        # Prepare shm segment for keys.
        keys_shm = shared_memory.SharedMemory(
            size=np.product(keys_shape) * dtypes.keys_dtype_size,
            create=True
        )
        keys_interface = np.ndarray(
            shape=keys_shape,
            dtype=dtypes.keys_dtype,
            buffer=keys_shm.buf
        )

        # Prepare shm segment for values.
        values_shm = shared_memory.SharedMemory(
            size=np.product(values_shape) * dtypes.map_dtype_size,
            create=True
        )
        values_interface = np.ndarray(
            shape=values_shape,
            dtype=dtypes.map_dtype,
            buffer=values_shm.buf
        )

        if config is None:
            config = load_database_config(out_dir)

        db = TablesDb(config.database_file)

        #
        print("Loading database keys...")
        db.read_keys(arr=keys_interface)

        #
        print("Loading database values...")
        db.read_values(arr=values_interface)

        db.close()

        # Remember the names for attaching in future.
        self.keys_shm_name = keys_shm.name
        self.values_shm_name = values_shm.name

        # Remember array data.
        self.keys_meta = {
            "shape": keys_shape,
            "dtype": dtypes.keys_dtype
        }
        self.values_meta = {
            "shape": values_shape,
            "dtype": dtypes.map_dtype
        }

        # Close my access.
        keys_shm.close()
        values_shm.close()




import os
import re
import time
import requests

from expam.classify import _STEP, ENTREZ_API, ENTREZ_EMAIL, ENTREZ_TOOL
from expam.database import FileLocationConfig
from expam.database.config import validate_taxonomy_files
from expam.utils import yield_csv


class TaxonomyNCBI:
    def __init__(self, file_config: FileLocationConfig) -> None:
        self.config: FileLocationConfig = file_config

    def find_downloaded_taxonomy(self):
        if not validate_taxonomy_files(self.config):
            raise OSError("Taxonomy files not located!")

    def load_taxonomy_map(self, convert_to_name=True):
        self.find_downloaded_taxonomy()

        # Create map from scientific name --> (taxid, rank).
        taxon_data = {}
        for data in yield_csv(self.config.taxon_rank):
            taxon_data[",".join(data[0:-2])] = tuple(data[-2:])

        # Create map from tax_id --> lineage (tuple).
        tax_id_to_lineage = {}
        for data in yield_csv(self.config.taxid_lineage):
            tax_id_to_lineage[data[0]] = tuple(data[1:])

        if not convert_to_name:
            return tax_id_to_lineage, taxon_data

        # Create map from name --> lineage (tuple).
        name_to_lineage = {}
        for data in yield_csv(self.config.accession_id):
            name_to_lineage[data[0]] = tax_id_to_lineage[data[2]]

        return name_to_lineage, taxon_data

    def load_sequence_map(self):
        return list(yield_csv(self.config.accession_id))

    def load_taxid_lineage_map(self):
        if os.path.exists(self.config.taxid_lineage):
            return list(yield_csv(self.config.taxid_lineage))
        else:
            return []

    def load_rank_map(self):
        name_to_rank = {}
        
        if os.path.exists(self.config.taxon_rank):
            for data in yield_csv(self.config.taxon_rank):
                if len(data) > 1:
                    name_to_rank[data[0]] = ",".join(data[1:])
        
        return name_to_rank

    def accession_to_taxonomy(self):
        """
        Map accession IDs to taxonomic labels.

        :sequence_ids: List - (sequence_id, accession_id, taxon_id)
        :taxon_ranks: List - (taxon_id, taxonomic ranks)
        :taxa_to_rank: Dict - taxon --> rank
        """

        def tuples_to_disk(lst):
            return "\n".join([",".join(item) for item in lst])

        def dict_to_disk(dct):
            return "\n".join([",".join((key, value)) for key, value in dct.items()])

        sequence_ids = self.load_sequence_map()
        taxon_ranks = self.load_taxid_lineage_map()
        taxa_to_rank = self.load_rank_map()

        # Collect taxon ids for unknown organisms.
        accessions_to_be_mapped = []
        taxa_to_be_collected = []

        for (sequence_id, accession_id, taxon_id) in sequence_ids:
            if taxon_id == "None":
                accessions_to_be_mapped.append(accession_id)
            else:
                taxa_to_be_collected.append(taxon_id)

        requestor = EntrezRequestor()

        if accessions_to_be_mapped:
            accession_to_tax = requestor.request_tax_ids("nuccore", accessions_to_be_mapped)

            print("Received %d response(s) for ESummary TaxID request!"
                % len(accession_to_tax))

            for i in range(len(sequence_ids)):
                if sequence_ids[i][1] in accession_to_tax:
                    sequence_ids[i][2] = accession_to_tax[sequence_ids[i][1]]
                    taxa_to_be_collected.append(sequence_ids[i][2])

        # Collect taxonomic lineages for taxa.
        current_taxa = {taxa[0] for taxa in taxon_ranks}
        taxa_to_be_collected = {  # Set so that we collect unique values.
            taxon_id
            for taxon_id in taxa_to_be_collected
            if taxon_id not in current_taxa
        }

        if taxa_to_be_collected:
            taxid_to_taxon, taxon_to_rank = requestor.request_labels("taxonomy", "xml", list(taxa_to_be_collected))

            print("Received %d response(s) for EFetch Taxon request!"
                % len(taxid_to_taxon))

            taxon_ranks.extend(taxid_to_taxon)
            taxa_to_rank.update(taxon_to_rank)

        # Save update maps to disk.
        with open(self.config.accession_id, "w") as f:
            f.write(tuples_to_disk(sequence_ids))

        with open(self.config.taxid_lineage, "w") as f:
            f.write(tuples_to_disk(taxon_ranks))

        print("Taxonomic lineages written to %s!" % self.config.taxid_lineage)

        with open(self.config.taxon_rank, "w") as f:
            f.write(dict_to_disk(taxa_to_rank))

        print("Taxonomic ranks written to %s!" % self.config.taxon_rank)


class EntrezRequestor:
    def __init__(self, entrez_tool: str = None, entrez_email: str = None, api_key: str = None) -> None:
        self.entrez_tool = ENTREZ_TOOL if entrez_tool is None else entrez_tool
        self.entrez_email = ENTREZ_EMAIL if entrez_email is None else entrez_email
        self.api_key = ENTREZ_API if api_key is None else api_key

    def request_tax_ids(self, db, id_list):
        POST_URL = "https://eutils.ncbi.nlm.nih.gov/entrez/eutils/esummary.fcgi"
        taxids, upto = {}, 0

        PARAMS = {
            "tool": self.entrez_tool,
            "email": self.entrez_email,
            "api_key": self.api_key,
            "db": db,
        }

        while upto < len(id_list):
            next_requests = id_list[upto:upto + _STEP]

            print("Posting %d UIDs to NCBI Entrez %s."
                % (len(next_requests), db))

            PARAMS["id"] = ",".join(next_requests)
            esummary_request = requests.post(
                url=POST_URL,
                data=PARAMS
            )

            # Parse TaxonIDs from raw results.
            accn_id = tax_id = None

            for line in esummary_request.text.split("\n"):
                if "<DocSum>" in line:  # Start new record.
                    accn_id = tax_id = None

                elif 'AccessionVersion' in line:
                    accn_id = self.parse_id(line)

                elif 'TaxId' in line:
                    tax_id = self.parse_tax_id(line)

                elif "</DocSum>" in line:  # Only save complete records.
                    if accn_id is not None and tax_id is not None:
                        taxids[accn_id] = tax_id

            upto += _STEP

            time.sleep(1.0)  # Allow server time to breath.

        return taxids

    @staticmethod
    def parse_id(string):
        new_id = re.findall(r'\<Item Name\="AccessionVersion" Type\="String"\>(.*?)\<\/Item\>', string)

        if not new_id:
            raise ValueError("No taxids found!")
        else:
            return new_id[0]

    @staticmethod
    def parse_tax_id(string):
        taxids = re.findall(r'\<Item Name\="TaxId" Type\="Integer"\>(.*?)\<\/Item\>', string)

        if not taxids:
            raise ValueError("No taxids found!")
        else:
            return taxids[0]

    def request_labels(self, db, retmode, id_list):
        POST_URL = "https://eutils.ncbi.nlm.nih.gov/entrez/eutils/efetch.fcgi"
        taxon_labels, ranks, upto = [], {}, 0

        PARAMS = {
            "tool": self.entrez_tool,
            "email": self.entrez_email,
            "api_key": self.api_key,
            "db": db,
            "retmode": retmode,
        }

        while upto < len(id_list):
            next_requests = id_list[upto:upto + _STEP]

            print("Posting %d UIDs to NCBI Entrez %s."
                % (len(next_requests), db))

            PARAMS["id"] = ",".join(next_requests)
            efetch_request = requests.post(
                url=POST_URL,
                data=PARAMS
            )

            # Parse taxonomic labels from raw results.
            tax_id = taxa = rank = name = None
            sub_tax_id = sub_rank = sub_name = None
            collect = True

            for line in efetch_request.text.split("\n"):
                if "<Taxon>" in line and collect:  # Start new record.
                    tax_id = taxa = name = rank = None

                elif "<Taxon>" in line and not collect:
                    sub_tax_id = sub_rank = sub_name = None

                elif "<TaxId>" in line and collect:
                    tax_id = self.parse_single_tax_id(line)

                elif "<TaxId>" in line and not collect:
                    sub_tax_id = self.parse_single_tax_id(line)

                elif "<Lineage>" in line and collect:
                    taxa = self.parse_lineage(line).replace(",", "")

                elif "<Rank>" in line and not collect:
                    if sub_name == "cellular organisms":
                        sub_rank = "top"

                    else:
                        sub_rank = self.parse_rank(line)

                elif "<Rank>" in line and collect:
                    if name == "cellular organisms":
                        rank = "top"

                    else:
                        rank = self.parse_rank(line)

                elif "<ScientificName>" in line and not collect:
                    sub_name = self.parse_name(line)

                elif "<ScientificName>" in line and collect:
                    name = self.parse_name(line)

                elif "<LineageEx>" in line:
                    collect = False

                elif "</Taxon>" in line and not collect:
                    ranks[sub_name] = sub_tax_id + "," + sub_rank

                elif "</LineageEx>" in line:
                    collect = True

                elif "</Taxon>" in line and collect:
                    if tax_id is not None and taxa is not None:
                        lineage = taxa.strip().split("; ")
                        if name not in lineage and name is not None:
                            lineage += [name]

                        taxon_labels.append([tax_id, ",".join(lineage)])

                    ranks[name] = tax_id + ',' + rank

            upto += _STEP

            time.sleep(1.0)  # Allow server time to breath.

        return taxon_labels, ranks

    @staticmethod
    def parse_single_tax_id(string):
        taxids = re.findall(r'\<TaxId\>(.*?)\<\/TaxId\>', string)

        if not taxids:
            raise ValueError("No taxids found!")
        else:
            return taxids[0]

    @staticmethod
    def parse_lineage(string):
        lineage = re.findall(r'\<Lineage\>(.*?)\<\/Lineage\>', string)

        if not lineage:
            raise ValueError("No lineages found!")
        else:
            return lineage[0]

    @staticmethod
    def parse_rank(string):
        rank = re.findall(r'\<Rank\>(.*?)\<\/Rank\>', string)

        if not rank:
            raise ValueError("No rank found!")
        else:
            return rank[0]

    @staticmethod
    def parse_name(string):
        name = re.findall(r'\<ScientificName\>(.*?)\<\/ScientificName\>', string)

        if not name:
            raise ValueError("No rank found!")
        else:
            name = re.sub(r"[,]", "", name[0])
            return name






from multiprocessing import shared_memory
import os
import queue
import re
import numpy as np
from expam.database import CHUNK_SIZE, EXPAM_TEMP_EXT, NULL_VALUE, TIMEOUT, DataTypeConfig, expam_dtypes
from expam.ext.kmers import get_raw_kmers
from expam.ext.map import classify
from expam.process import COMMAND_CODES
from expam.process.jobworker import JobWorker


class ClassifyWorker(JobWorker):
    def __init__(
        self, k, out_dir, temp_dir, kmer_db, lca_matrix, n_readers, logging_dir, alpha, status_index,
        statuses, child_statuses, timeout=TIMEOUT, phases=None, dtype_config: DataTypeConfig = expam_dtypes
    ):
        super(ClassifyWorker, self).__init__(logging_dir, status_index, statuses, child_statuses, timeout=timeout,
                                             phases=phases)
        self.job_name = "classify"
        self.n_readers = n_readers

        # Configuration parameters.
        self.k = k
        self.alpha = alpha
        self.chunks = (k // CHUNK_SIZE) + (k % CHUNK_SIZE != 0)

        self.out_dir = out_dir
        self.temp_dir = temp_dir

        self._read_array_length = 150
        self.read_array = np.ndarray((self._read_array_length, self.chunks + 1), dtype=dtype_config.keys_dtype)

        self.RAW = "_%s.reads_%s" % ("%d", EXPAM_TEMP_EXT)

        # Load the database.
        self.dtypes: DataTypeConfig = dtype_config
        self.keys_shm = shared_memory.SharedMemory(
            name=kmer_db.keys_shm_name,
            create=False
        )
        self.keys = np.ndarray(
            shape=kmer_db.keys_meta["shape"],
            dtype=kmer_db.keys_meta["dtype"],
            buffer=self.keys_shm.buf
        )
        self.values_shm = shared_memory.SharedMemory(
            name=kmer_db.values_shm_name,
            create=False
        )
        self.values = np.ndarray(
            shape=kmer_db.values_meta["shape"],
            dtype=kmer_db.values_meta["dtype"],
            buffer=self.values_shm.buf
        )

        self.current_file_data = [("", []) for _ in range(n_readers)]

        # LCA Processing.
        self.lca_matrix = lca_matrix

        self.init_methods_from_dict(
            {
                "classify": {
                    "getter": None,
                    "processor": self.classify_read,
                    "transition": self.save_final_data,
                }
            }
        )

    def classify_read(self, read_data):
        new_file_name, read_id, read, reader_id = read_data

        # Continue to get reads without checking for errors; only check for errors when there is a break in workflow.
        while True:
            read_id = read_id.decode("utf8")

            max_n_kmers = len(read) - self.k + 1
            if max_n_kmers > self.read_array.shape[0]:
                self.read_array = np.ndarray((max_n_kmers, self.chunks + 1), dtype=self.dtypes.keys_dtype)

            if new_file_name != self.current_file_data[reader_id][0]:
                if len(self.current_file_data[reader_id][1]) > 0:
                    self.save_temp_data(reader_id)

                self.current_file_data[reader_id] = (new_file_name, [])

            result, best_clade, read_length, result_string = self.process_read(read)
            self.current_file_data[reader_id][1].append(
                [result, read_id, "p" + str(best_clade), str(read_length), result_string]
            )

            # Get a head start on any available work.
            try:
                command_code, data = self.work_queue.get(block=True, timeout=self._timeout)

                if command_code != COMMAND_CODES["ACTIVE"]:
                    self.work_queue.put((command_code, data))
                    break

                new_file_name, read_id, read, reader_id = data
            except queue.Empty:
                break

    def save_temp_data(self, reader_id):
        def format_result(result_list):
            return "\n".join(["\t".join(item) for item in result_list])

        def base_name(file_url):
            return re.match(r"(\S+)\.(?:fa|fq|fasta|fastq)(?:\.tar)*(?:\.gz)*$", file_url).group(1)

        file_name, file_data = self.current_file_data[reader_id]

        my_file_name = os.path.join(self.temp_dir, base_name(file_name) + (self.RAW % self.pid))
        with open(my_file_name, "a+") as f:
            f.write(format_result(file_data))

    def process_read(self, read):
        read_length = len(read)
        n_kmers = get_raw_kmers(read, self.k, self.read_array)

        # Control for reads that don't have long enough good regions for kmers to be extracted.
        if n_kmers == 0:
            return "U", 0, read_length, ""

        # Map the kmers.
        cls, common_clade, kmer_string = classify(self.read_array[:n_kmers], self.keys, self.values,
                                                  self.lca_matrix, NULL_VALUE, self.alpha)

        if cls == -1:
            result = "U"

        elif cls == 1:
            result = "S"

        else:
            result = "C"

        return result, common_clade, read_length, kmer_string

    def save_final_data(self, *args):
        for reader_id in range(self.n_readers):
            if len(self.current_file_data[reader_id][1]) > 0:
                self.save_temp_data(reader_id)

        self.keys_shm.close()
        self.values_shm.close()



RANK_ORDER = [
    "0", "root", "top", "superkingdom", "kingdom", "subkingdom", "superclade", "clade", "subclade",
    "superphylum", "phylum", "subphylum", "superclass", "class", "subclass", "superorder", "order",
    "suborder", "superfamily", "family", "subfamily", "supergenus", "genus", "subgenus", "species group",
    "species subgroup", "superspecies", "species", "subspecies", "serogroup", "serotype", "superstrain",
    "strain", "substrain", "no rank"
]
TAX_RESULTS_HEADER = [
    "taxid", "%ClassifiedCumulative", "ClassifiedCumulative", "ClassifiedCount",
    "%SplitCumulative", "SplitCumulative", "SplitCount", "Rank", "Lineage"
]

PHYLA_COLOURS = [
    ("Proteobacteria", "#E85343"),
    ("Bacteroidetes", "#30B674"),
    ("Firmicutes", "#6585C3"),
    ("Actinobacteria", "#FEDB45")
]



import multiprocessing
import queue
import numpy as np
from expam.database import CHUNK_SIZE, TIMEOUT, DataTypeConfig, expam_dtypes
from expam.ext.kmers import get_raw_kmers, remove_duplicates, to_32bit, import_mask, map_kmers
from expam.ext.map import binarysearch
from expam.logger import log_timer
from expam.process import COMMAND_CODES
from expam.process.jobworker import JobWorker
from expam.sequences import MaskStore, SequenceStore


class ExtractWorker(JobWorker):
    def __init__(
        self, k, shm_kmer_params, kmer_ranges, node_map, out_dir, logging_dir,
        status_index, statuses, child_statuses, timeout=TIMEOUT,
        phases=None, dtype_config: DataTypeConfig = expam_dtypes
    ):
        super(ExtractWorker, self).__init__(logging_dir, status_index, statuses, child_statuses,
                                            timeout=timeout, phases=phases)

        self.job_name = "extract"
        self.k = k
        self.kmer_ranges = kmer_ranges

        self.chunks = (k // CHUNK_SIZE) + (k % CHUNK_SIZE != 0)
        self.keys_dtype = dtype_config.keys_dtype
        self.d_type = np.dtype(
            [
                ('index', self.keys_dtype),
                *(
                    ('k%d' % chunk, self.keys_dtype)
                    for chunk in range(self.chunks)
                )
            ]
        )

        # For mapping LCAs.
        self.node_map = node_map

        # Hold the sequences and masks that I have extracted.
        self.sequence_store = SequenceStore(k, out_dir)
        self.mask_store = MaskStore()

        shm_kmer_name, shm_kmer_shape = shm_kmer_params
        allocation_shape = (shm_kmer_shape[0], shm_kmer_shape[1] + 1)

        # Long-term arrays used for extraction.
        self.allocation = np.zeros(
            shape=allocation_shape,  # Account for index col.
            dtype=self.keys_dtype,
            order="C",  # Row-major storage (consecutive k-mer access).
        )

        # Long-term array for sending kmers to children.
        self.out_shape = shm_kmer_shape
        self.out_name = shm_kmer_name
        self.out_allocation = multiprocessing.shared_memory.SharedMemory(shm_kmer_name)
        self.out_array = np.ndarray(
            shape=shm_kmer_shape,
            dtype=self.keys_dtype,
            buffer=self.out_allocation.buf
        )

        # Primary queue for getting work during mapping phase.
        self.map_queue = multiprocessing.Queue()

        # Initialise methods for execution.
        self.init_methods_from_dict(
            {
                "import": {
                    "getter": None,
                    "processor": self.import_sequence,
                    "transition": self.prepare_map,
                },
                "map": {
                    "getter": self.get_my_sequence,
                    "processor": self.send_for_mapping,
                    "transition": self.empty_queue,
                },
                "shutdown": self.close_shm
            }
        )

    @log_timer
    def import_sequence(self, sequence_url):
        def resize_base(size):
            nonlocal base

            base = base[:size]
            update_view()

        def new_view():
            nonlocal base
            return base.ravel().view(dtype=self.d_type)

        def update_view():
            nonlocal base, view
            view = base.ravel().view(dtype=self.d_type)

        # Collect raw sequence.
        self.log("Importing sequence at %s." % sequence_url)
        seq_id, seq_length = self.sequence_store.add_sequence(sequence_url)

        # Get view on required segment of allocation.
        base = self.allocation[:seq_length]
        view = new_view()

        # Create draft mask and get corresponding kmers.
        new_size = get_raw_kmers(
            sequence=self.sequence_store[seq_id],
            k=self.k,
            arr=base
        )
        resize_base(new_size)

        # Sort by kmers, which also sorts the indices
        view.sort(
            order=['k%d' % chunk for chunk in range(self.chunks)],
            kind="stable"
        )
        self.log("Removing duplicates.")
        new_size = remove_duplicates(base)
        resize_base(new_size)

        # Save mask as 32-bit array.
        self.mask_store[seq_id] = to_32bit(view['index'])
        self.send_kmers(
            ranges=self.kmer_ranges,
            kmers=base[:, 1:]
        )

    @staticmethod
    def clear_temp_arrays(arr):
        arr[:] = 0

    @log_timer
    def send_kmers(self, ranges, kmers, node_id=None):
        target = np.ndarray(shape=(kmers.shape[1]), dtype=kmers.dtype)
        target[:] = ranges[0][0]

        total = 0
        n = len(ranges) - 1
        lower_ind = upper_ind = self.get_index(target, kmers, greater=False)

        # Wait for previous data to be sent.
        while self.loop_condition(not np.all(self.out_array == 0)):
            self.timeout()

        # Place kmers in out_array.
        self.out_array[:kmers.shape[0], :] = kmers[:]

        # Separate data between children.
        for i, (lower_bound, upper_bound) in ranges.items():
            target[:] = upper_bound
            upper_ind = self.get_index(target, kmers)

            # Upper bound can only attain equality for maximum hash value.
            if i == n:
                while (upper_ind < kmers.shape[0]) and self.kmers_equal(target, kmers[upper_ind]):
                    upper_ind += 1

            # Don't send empty work.
            if lower_ind != upper_ind:

                message_data = (
                    (lower_ind, upper_ind),
                    (self.out_name, self.out_shape),
                )
                if node_id is not None:
                    message_data += (node_id,)

                self.tell_child(
                    cid=i,
                    data=(
                        COMMAND_CODES["ACTIVE"],
                        message_data
                    )
                )

                total += upper_ind - lower_ind

            # Beginning next search.
            lower_ind = upper_ind

        self.log("%d / %d kmers sent!" % (total, kmers.shape[0]))

    def get_index(self, target, kmers, greater=True):
        def _check_ind():
            nonlocal ind, kmers

            if (greater and ind == kmers.shape[0]) or ((not greater) and ind == 0):
                return True

            return False

        # Find closest element in list. Note that this element
        # may be less than the bound.
        _, ind = binarysearch(target, kmers)

        if _check_ind():
            return int(ind)

        while (not _check_ind()) and self.kmers_compare(kmers[ind], target, greater):
            if greater:
                ind += 1

            else:
                ind -= 1

        return int(ind)

    @staticmethod
    def kmers_equal(one, two):
        return np.all(one == two)

    @staticmethod
    def kmers_compare(one, two, greater):
        for i in range(one.shape[0]):
            if one[i] < two[i]:
                return greater

            elif one[i] > two[i]:
                return not greater

            elif i + 1 == one.shape[0]:
                return not greater

    def prepare_map(self, *args):
        # Save accession ids from my sequences.
        self.sequence_store.save_accession_ids(suffix=str(self._status_id))

        # Prepare mapping jobs.
        self.load_queue()

    def load_queue(self):
        for seq_id in self.sequence_store.keys():
            self.map_queue.put(
                (
                    COMMAND_CODES["ACTIVE"],
                    seq_id
                )
            )

    def get_my_sequence(self, *args):

        # Try to get local work first.
        try:
            work = self.map_queue.get(block=True, timeout=self._timeout)

            return work

        except queue.Empty:
            pass

        # Now look for commands from parent.
        try:
            work = self.work_queue.get(block=True, timeout=self._timeout)

            return work

        except queue.Empty:
            return None

    def send_for_mapping(self, seq_id):
        node_id = self.node_map[seq_id]

        sequence = self.sequence_store[seq_id]
        mask = self.mask_store[seq_id]
        n_kmers = mask.shape[0]

        # Get view on required segment of allocation.
        base = self.allocation[:n_kmers]
        view = base.ravel().view(dtype=np.dtype(
            [
                ('index', self.keys_dtype),
                *(
                    ('k%d' % chunk, self.keys_dtype)
                    for chunk in range(self.chunks)
                )
            ]
        ))

        # Place 32-bit mask in 64-bit array.
        import_mask(
            data=mask,
            dest=view['index']
        )
        map_kmers(
            sequence=sequence,
            k=self.k,
            arr=base
        )

        # Send these kmers to be mapped.
        self.send_kmers(
            ranges=self.kmer_ranges,
            kmers=base[:, 1:],
            node_id=node_id
        )

    def empty_queue(self, *args):
        try:
            _ = self.map_queue.get(block=True, timeout=self._timeout)

            raise ValueError("Mapping queue was not empty!")

        except queue.Empty:

            return

    def close_shm(self):
        try:
            self.out_allocation.unlink()
            self.out_allocation.close()

        except FileNotFoundError:
            pass




"""

███████╗██╗░░██╗██████╗░░█████╗░███╗░░░███╗
██╔════╝╚██╗██╔╝██╔══██╗██╔══██╗████╗░████║
█████╗░░░╚███╔╝░██████╔╝███████║██╔████╔██║
██╔══╝░░░██╔██╗░██╔═══╝░██╔══██║██║╚██╔╝██║
███████╗██╔╝╚██╗██║░░░░░██║░░██║██║░╚═╝░██║
╚══════╝╚═╝░░╚═╝╚═╝░░░░░╚═╝░░╚═╝╚═╝░░░░░╚═╝


Metagenomic analysis using a reference phylogeny.


Uploading to PyPi
=================

Run ./buildandtwine.sh Shell script to build source code and upload.


Uploading to bioconda
=====================

Create skeleton recipe from PyPi repo.
    conda skeleton pypi --extra-specs numpy --extra-specs Cython expam


"""

import os
from setuptools import setup, find_packages
from setuptools.extension import Extension
import sys

from Cython.Build import cythonize
import numpy as np

SOURCE = os.path.dirname(os.path.abspath(__file__))
sys.path.append(SOURCE)

from src.expam import __version__


# Get project description.
with open(os.path.join(SOURCE, "README.md"), mode="r", encoding="utf-8") as f:
    long_description = f.read()

# Extension instances for Cython scripts.
extensions = [
    *cythonize(
        Extension(
            "kmers._build.kmers",
            sources=["src/expam/ext/kmers/extract.pyx", "src/expam/ext/kmers/kmers.c", "src/expam/ext/kmers/jellyfish.c"],
            include_dirs=[np.get_include()],
            extra_compile_args=["-std=c99"],
            define_macros=[("NPY_NO_DEPRECATED_API", "NPY_1_7_API_VERSION")]
        ),
        language_level="3",
        build_dir="src/expam/ext/kmers/_build"
    ),
    *cythonize(
        Extension(
            "map._build.map",
            sources=["src/expam/ext/map/map.pyx"],
            include_dirs=[np.get_include()],
            define_macros=[("NPY_NO_DEPRECATED_API", "NPY_1_7_API_VERSION")]
        ),
        language_level="3",
        build_dir="src/expam/ext/map/_build"
    ),
    *cythonize(
        Extension(
            "sets._build.sets",
            sources=["src/expam/ext/sets/sets.pyx", "src/expam/ext/sets/mfil.c"],
            include_dirs=[np.get_include()],
            extra_compile_args=["-std=c99"],
            define_macros=[("NPY_NO_DEPRECATED_API", "NPY_1_7_API_VERSION")]
        ),
        language_level="3",
        build_dir="src/expam/ext/sets/_build"
    )
]

setup(
    #
    # Metadata.
    #
    name="expam",
    version="%d.%d.%d.%d" % __version__,
    description="Metagenomic profiling using a reference phylogeny",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/seansolari/expam",
    author="Sean Solari",
    author_email="sean.solari@monash.edu",
    classifiers=[
        "Development Status :: 3 - Alpha",
        "Intended Audience :: Science/Research",
        "Intended Audience :: Developers",
        "Topic :: Scientific/Engineering :: Bio-Informatics",
        "License :: OSI Approved :: BSD License",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
    ],
    #
    # Source files.
    #
    package_dir={'': 'src'},
    packages=find_packages(where='src'),
    #
    # External dependencies.
    #
    python_requires=">=3.8",
    install_requires=[
        "numpy>=1.22.0",
        "matplotlib",
        "multiprocess",
        "pandas",
        "psutil",
        "requests",
        "tables"
    ],
    extras_require={
        "treeplot": ["ete3", "PyQt5==5.12", "sourmash"]
    },
    #
    # Cython modules.
    #
    ext_package="expam.ext",
    ext_modules=extensions,
    #
    # Make main callable from console.
    #
    entry_points={
        "console_scripts": [
            "expam=expam.main:main",
            "expam_limit=expam.sandbox.main:main"
        ],
    },
    #
    # Source links.
    #
    project_urls={
        "Bug Reports": "https://github.com/seansolari/expam/issues",
    },
)



from expam.tree.location import Location
from expam.tree.tree import Index


def simulate_balanced_phylogeny(names):
    """
    :param names: List - list of leaf names.
    """

    # Declare an index.
    index = Index()

    # Insert all children.
    for name in names:
        # Make child node.
        childNode = Location(
            name=name,
            type="Leaf",
            dist=1.0
        )
        # Insert into tree.
        index.append(childNode)

    # Naming string.
    _NAME = len(names) - 1

    # Repeatedly join pairs until a full tree is made.
    while len(names) > 1:
        new_names = []

        while len(names) > 0:
            try:
                # Get next two children.
                child_names = []
                i = 0
                while i < 2:
                    child_names.append(names.pop())
                    i += 1

            except IndexError:
                # Only one child left. Just pass on the name.
                new_names.append(child_names[0])
                break

            # Join these children:
            # Make a parent node.
            parentName = str(_NAME)
            _NAME -= 1
            # Make a new Location.
            parentNode = Location(
                name=parentName,
                type="Branch"
            )
            print(f'New Parent {parentName}...')
            # Apply join.
            index.join(
                child_names[0],
                child_names[1],
                parentNode
            )

            # Append parent name to new_names.
            new_names.append(parentName)

        # New lot of children.
        names = new_names

    # Convert tree to newick format.
    return index.to_newick() + ";"



class Location:
    """Represents a node in the phylogeny."""
    is_root = False

    def __init__(self, name="", type="", dist=0.0, coord=None, accession_id=None, taxid=None, **kwargs):
        """__init__ 

        :param name: name of node, defaults to ""
        :type name: str, optional
        :param type: Leaf or Branch, defaults to ""
        :type type: str, optional
        :param dist: distance to parent node, defaults to 0.0
        :type dist: float, optional
        :param coord: binary coordinate from root to node, defaults to None
        :type coord: list, optional
        :param accession_id: NCBI accession id, defaults to None
        :type accession_id: str, optional
        :param taxid: NCBI taxonomy id, defaults to None
        :type taxid: int, optional
        :ivar name: node name
        :ivar type: "Leaf" or "Branch"
        :ivar distance: distance to parent node
        :ivar coordinate: list of binary binary numbers representing path from root to node
        :ivar nchildren: number of children below this node
        :ivar accession_id: NCBI accession id (only valid for leaves)
        :ivar taxid: NCBI taxonomy id
        """
        self._name = name
        self._type = type
        self._distance = dist
        self._coordinate = [] if coord is None else coord
        self._nchildren = 0
        self._accession_id = accession_id
        self._taxid = taxid

        self.children = []
        self.primaryChild = None

    def __str__(self):
        return (f"Location - {self._name} ({self._type}):"
                f"\n\t+ Coordinate: {self._coordinate}"
                f"\n\t+ Children: {self._nchildren}")

    def name():
        doc = "The name property."

        def fget(self):
            return self._name

        def fset(self, value):
            self._name = value

        def fdel(self):
            del self._name

        return locals()

    name = property(**name())

    def type():
        doc = "The type property."

        def fget(self):
            return self._type

        def fset(self, value):
            self._type = value

        def fdel(self):
            del self._type

        return locals()

    type = property(**type())

    def distance():
        doc = "The distance property."

        def fget(self):
            return self._distance

        def fset(self, value):
            self._distance = value

        def fdel(self):
            del self._distance

        return locals()

    distance = property(**distance())

    def coordinate():
        doc = "The coordinate property."

        def fget(self):
            return self._coordinate

        def fset(self, value):
            self._coordinate = value

        def fdel(self):
            del self._coordinate

        return locals()

    coordinate = property(**coordinate())

    def nchildren():
        doc = "The nchildren property."

        def fget(self):
            return self._nchildren

        def fset(self, value):
            self._nchildren = value

        def fdel(self):
            del self._nchildren

        return locals()

    nchildren = property(**nchildren())

    def accession_id():
        doc = "The accession id property."

        def fget(self):
            return self._accession_id

        def fset(self, value):
            self._accession_id = value

        def fdel(self):
            del self._accession_id

        return locals()

    accession_id = property(**accession_id())

    def taxid():
        doc = "The accession id property."

        def fget(self):
            return self._taxid

        def fset(self, value):
            self._taxid = value

        def fdel(self):
            del self._taxid

        return locals()

    taxid = property(**taxid())

    @classmethod
    def make_branch(cls, branch_name):
        return Location(branch_name, "Branch", 0)

    @classmethod
    def make_leaf(cls, tree, leaf_name, dist):
        return Location(leaf_name, "Leaf", dist)

    def add_child(self, child, primary_child=False):
        self.children.__setitem__(child.name, child)

        if primary_child:
            self.set_primary(child)

    def set_primary(self, node):
        self.primaryChild = node




import random
from math import floor, log
import os
import re
import sys
import traceback
import pandas as pd
from expam.tree import PHYLA_COLOURS
from expam.tree.location import Location


def propose_lca(c1, c2):
    """
    Given two binary coordinates, compute the shortest common prefix.

    :param coord: Indexable.
    :return: List.
    """
    m = min(len(c1), len(c2))
    if m == 0:
        return []

    for i in range(1, m + 1):
        if c1[-i] != c2[-i]:
            if i == 1:
                return []
            else:
                return c1[-i + 1:]
    else:
        return c1[-m:]


class Index:
    """ Phylogeny index that can load, save and manipulate Newick trees.
    """
    def __init__(self):
        self.db = None
        self.leaf_taxid = None
        self.taxid_lineage = None

        self._pointers = {}  # Pointers is a map from node name to pool location.
        self.pool = [Location()]  # Empty location to represent an 'unclassified' location.

    def __len__(self):
        return len(self.pool)

    def __repr__(self):
        return self.__str__()

    def __str__(self):
        return f'<Phylogeny Index, length={len(self)}>'
        
    def use_db(self, db_path: str):
        if not os.path.exists(db_path):
            raise OSError("Can't find database path %s." % db_path)

        phy_path = os.path.join(db_path, 'phylogeny')
        accession_ids_path = os.path.join(phy_path, 'accession_ids.csv')
        taxid_lineage_path = os.path.join(phy_path, 'taxid_lineage.csv')

        for path in (phy_path, accession_ids_path, taxid_lineage_path):
            if not os.path.exists(path):
                raise OSError("Can't find required data %s. Have you run download_taxonomy?")

        self.load_leaf_taxid(accession_ids_path)
        self.load_taxid_lineage(taxid_lineage_path)

    def load_leaf_taxid(self, accession_ids_path: str):
        if self.leaf_taxid is None:
            self.leaf_taxid = {}
        else:
            return

        with open(accession_ids_path, 'r') as f:
            for line in f:
                name, _, taxid = line.strip().split(',')
                self.leaf_taxid[name] = taxid

    def load_taxid_lineage(self, taxid_lineage_path: str):
        if self.taxid_lineage is None:
            self.taxid_lineage = {}
        else:
            return

        with open(taxid_lineage_path, 'r') as f:
            for line in f:
                data = line.strip().split(',')
                self.taxid_lineage[data[0]] = tuple(data[1:])

    def get_node_lineages(self, node):
        taxids = [self.leaf_taxid[leaf] for leaf in self.get_child_leaves(node)]
        return [self.taxid_lineage[taxa] for taxa in taxids]

    @classmethod
    def load_newick(cls, path, keep_names=False, verbose=True):
        """load_newick Load Newick tree from file.

        :param path: path to Newick file
        :type path: str
        :raises OSError: file does not exist
        :return: name of leaves and phylogeny Index object
        :rtype: List[str], expam.tree.Index
        """
        newick_str = ""

        if not os.path.exists(path):
            raise OSError("Couldn't find path %s." % path)

        with open(path, "r") as f:
            for line in f:
                newick_str += line.strip()

        return cls.from_newick(newick_str, keep_names=keep_names, verbose=verbose)

    @classmethod
    def from_newick(cls, newick_string, keep_names=False, verbose=True):
        """from_newick Parse Newick string.

        :param newick_string: Newick string encoding tree.
        :type newick_string: str
        :return: name of leaves and phylogeny Index object
        :rtype: List[str], expam.tree.Index
        """
        stream = sys.stdout if verbose else open(os.devnull, 'w')

        # Remove whitespace.
        newick_string = newick_string.replace(" ", "").replace("\n", "")

        print("* Initialising node pool...", file=stream)
        index_pool = cls.init_pool(newick_string)

        print("* Checking for polytomies...", file=stream)
        cls.resolve_polytomies(index_pool)

        print("* Finalising index...", file=stream)
        leaf_names, index = cls.from_pool(index_pool, keep_names=keep_names)

        if stream != sys.stdout:
            stream.close()

        return leaf_names, index

    @staticmethod
    def init_pool(newick):
        pool = [Location()]

        i = 0
        parent_stack = []
        currentNode = None

        #
        # Create a new node.
        #
        def new_node(name, type):
            nonlocal currentNode

            currentNode = Location(name, type)

            pool.append(currentNode)

            # Top node has no ancestor.
            if len(parent_stack) > 0:
                parent_stack[-1].children.append(currentNode)

        #
        # Get node name.
        #
        def close_branch():
            nonlocal parent_stack, currentNode, i, newick, NEWICK_PARSE

            current_branch = parent_stack.pop()

            # Get this branch's name, if it is given.
            if newick[i + 1] not in NEWICK_PARSE:
                i += 1
                branch_name = parse_string()

                current_branch.name = branch_name

            currentNode = current_branch

        #
        # New branch in phylogeny and parent stack.
        #
        def new_branch():
            nonlocal currentNode

            new_node(name="", type="Branch")
            parent_stack.append(currentNode)

        #
        # New Leaf in phylogeny.
        #
        def new_leaf():
            nonlocal currentNode

            new_node(name="", type="Leaf")

        #
        # Parse name.
        #
        def parse_string(force_digits=False):
            nonlocal i  # Inherit current index.

            # Detect quotations.
            if newick[i] == "'":
                terminal = "'"

                i += 1

            else:
                terminal = NEWICK_PARSE

            j = i + 1
            while newick[j] not in terminal:
                j += 1

            string = newick[i:j]

            # Check formatting.
            if force_digits:
                try:
                    string = float(string)

                except ValueError:
                    raise ValueError("Invalid distance declaration %s!" % newick[i:j])

            if terminal == "'":
                i = j
            else:
                i = j - 1

            return string

        #
        # Clear current node (stack-ish).
        #
        def clear_node():
            nonlocal currentNode

            currentNode = None

        #
        # Set distance of current node.
        #
        def set_node_distance():
            nonlocal currentNode, i

            i += 1  # Skip ':' character.
            currentNode.distance = parse_string(force_digits=True)

        #
        # Do nothing.
        #
        def finisher():
            nonlocal i

            i = len(newick)

        # Parsing functions for Newick format.
        NEWICK_PARSE = {
            "(": new_branch,
            ")": close_branch,
            ",": clear_node,
            ":": set_node_distance,
            ";": finisher
        }

        # Parsing loop.
        while i < len(newick):
            if newick[i] in NEWICK_PARSE:
                NEWICK_PARSE[newick[i]]()

            else:
                new_leaf()

                currentNode.name = parse_string()

            i += 1

        # Return pool array for insertion into expam Index.
        return pool

    @staticmethod
    def resolve_polytomies(pool):
        """
        If the phylogeny contains polytomies, continually join the first two
        children with parents of distance 0 until the polytomy is resolved.

        :param pool: List.
        :return: None
        :rtype: None
        """
        # Processing loop.
        i = 1
        while i < len(pool):
            node = pool[i]

            # Continually join first two nodes.
            while len(node.children) > 2:
                print("\tPolytomy (degree=%d) detected! Resolving..."
                      % len(node.children))

                # Declare new parent of distance 0.
                newParent = Location(
                    name="",
                    type="Branch",
                    dist=0.0
                )

                # Replace children.
                for _ in range(2):
                    newParent.children.append(node.children.pop(0))

                # Insert into pool.
                node.children.append(newParent)
                pool.insert(i + 1, newParent)

            i += 1

    @classmethod
    def from_pool(cls, pool, keep_names=False, map_children_objects=True, update_nchildren=True):
        index = cls()
        index.pool = pool
        leaf_names = index.update_via_pool(keep_names=keep_names, map_children_objects=map_children_objects, update_nchildren=update_nchildren)

        return leaf_names, index

    def update_via_pool(self, keep_names=False, map_children_objects=True, update_nchildren=True):
        self._pointers = {}
        leaf_names = []

        branch_id = 1
        for i, node in enumerate(self.pool):
            if i == 0:
                continue

            if node.type == "Branch":
                if not keep_names:
                    node.name = str(branch_id)
                branch_id += 1
            else:
                # Store leaf names.
                leaf_names.append(node.name)

            # Set name mapping.
            self._pointers[node.name] = i

        if map_children_objects:
            # Replace children objects with respective names.
            # Note: This requires the name mapping to be complete...
            for node in self.pool[1:]:
                if node.type == "Branch":
                    node.children = [child.name for child in node.children]

        # Set (binary) coordinates.
        self.update_coordinates()

        if update_nchildren:
            # Update nchildren.
            self.update_nchildren()

        return leaf_names

    def append(self, item):
        # Declare an index.
        index = len(self.pool)

        # Set pointer.
        self._pointers[item.name] = index

        self.pool.append(item)

    def join(self, a, b, glue, dist=None):
        if dist is None:
            dist = [1.0, 1.0]
        i, j = self._pointers[a], self._pointers[b]
        i, j = self.smaller(i, j)  # i < j.

        one, two = j, self.neighbouring(j)
        three = self.neighbouring(i)
        self.move([one, two], three)

        self.insert(i, glue)

        # Update number of children and distances.
        glue.children.extend((a, b))

        for i, child in enumerate([a, b]):
            glue.nchildren += 1 + ((self[child].type == "Branch") * self[child].nchildren)

            self[child].distance = dist[i]

    def move(self, i, j):
        # NOTE: i > j.
        width = i[1] - i[0]

        if i[0] == j:
            return

        # Delete affected pointers.
        for n in range(j, i[1]):
            del self._pointers[n]

        # Retrieve items to be moved.
        temp = []
        for n in range(width):
            temp.append(self.pool.pop(i[1] - (n + 1)))

        # Insert these items in at j. Since they were popped in reverse order,
        # insert append them in the same index to revert back to the original
        # order.
        for obj in temp:
            self.pool.insert(j, obj)

        # Update pointers.
        for n in range(j, i[1]):
            self._pointers[n] = self.pool[n].name

    def insert(self, i, item):
        # Update pointers.
        for n in range(i, len(self.pool)):
            loc = self.pool[n]

            self._pointers[loc.name] = n + 1

        self.pool.insert(i, item)
        self._pointers[item.name] = i

    @staticmethod
    def _format_node_name(node_name):
        if node_name[0] == 'p':
            return str(node_name[1:])
        else:
            return str(node_name)

    def give_branch_name(self, node_name):
        node = self[node_name]
        if node.type == "Branch" and node.name[0] != 'p':
            return 'p' + node_name
        else:
            return node_name

    def __setitem__(self, key, value):
        try:
            pool_index = self._pointers[key]
        except KeyError:
            pool_index = self._pointers[self._format_node_name(key)]
        self.pool[pool_index] = value

    def __getitem__(self, key):
        try:
            pool_index = self._pointers[key]
        except KeyError:
            pool_index = self._pointers[self._format_node_name(key)]
        return self.pool[pool_index]

    def __delitem__(self, key):
        self.pool.pop(self._pointers[str(key)])

    def neighbouring(self, i):
        return i + 1 + self.pool[i].nchildren

    def update_coordinates(self):
        coord, i = [-1], 1

        while i < len(self.pool):
            while coord[0] + 1 > 1:
                coord.pop(0)

            coord[0] += 1

            if self.pool[i].type == "Branch":
                self.pool[i].coordinate = list(coord[:-1])

                coord.insert(0, -1)
            else:  # It's a leaf.
                self.pool[i].coordinate = list(coord[:-1])

            i += 1

    def update_nchildren(self):
        for node in self.pool[1:][::-1]:
            node.nchildren = len(node.children) + sum([
                self[child].nchildren
                for child in node.children
            ])

    def coord(self, coordinate):
        """coord Return Location (node) at coordinate.

        :param coordinate: binary list representing path to node
        :type coordinate: list
        :return: node in tree
        :rtype: expam.tree.Location
        """
        pool_index = 1

        for step in coordinate[::-1]:
            pool_index += 1
            pool_index += int(step) * (self.pool[pool_index].nchildren + 1)

        return self.pool[pool_index]

    def print_index(self):
        for l in self.pool:
            print(l.name, l.coordinate, l.nchildren)

    @staticmethod
    def smaller(a, b):
        if a < b:
            return a, b
        return b, a

    def to_newick(self):
        """to_newick Output tree to Newick format.

        :return: Newick format tree
        :rtype: str
        """
        if len(self.pool) == 2:
            return self.pool[1].name + ";"

        NODE_FORMAT = "$node$"

        newick = "(%s,%s)%s;" % (NODE_FORMAT, NODE_FORMAT, self.pool[1].name)
        node_formats = {
            "Branch": "(%s,%s){name}:{distance}" % (NODE_FORMAT, NODE_FORMAT),
            "Leaf": "{name}:{distance}"
        }

        for node in self.pool[2:]:
            node_data = node_formats[node.type].format(name=node.name, distance=node.distance)
            newick = newick.replace(NODE_FORMAT, node_data, 1)

        return newick

    def yield_child_nodes(self, node_name):
        """yield_child_nodes Yields node and children nodes (both branches and leaves).

        :param node_name: name of node to start yielding from
        :type node_name: str
        :yield: node names at or below node_name
        :rtype: str
        """
        node_name = self._format_node_name(node_name)
        active_nodes = [node_name]

        while active_nodes:
            node = active_nodes.pop(0)

            yield node
            active_nodes.extend(self[node].children)

    def yield_leaves(self, node_name):
        """yield_leaves Yield only the leaves at or below some node.

        :param node_name: node to retrieve leaves from.
        :type node_name: str
        :yield: leaf names at or below node_name.
        :rtype: str
        """
        node_name = self._format_node_name(node_name)

        if self[node_name].type == "Leaf":
            yield node_name
            return

        branches = [node_name]

        while branches:
            branch = branches.pop(0)

            for child in self[branch].children:
                if self[child].type == "Leaf":
                    yield child
                else:
                    branches.append(child)

    def get_child_nodes(self, node_name):
        """get_child_nodes Return list of nodes at or below node_name.

        :param node_name: name of node 
        :type node_name: str
        :return: list of node names
        :rtype: List[str]
        """
        return list(self.yield_child_nodes(node_name))

    def get_child_leaves(self, node_name):
        """get_child_leaves Get list of leaves at or below node_name.

        :param node_name: name of node
        :type node_name: str
        :return: list of leaf names
        :rtype: List[str]
        """
        return list(self.yield_leaves(node_name))

    def lca(self, *names):
        """lca Return name of the lowest common ancestor of these two nodes.

        Note that coordinates are read from right-to-left.

        :param name_one: name of node
        :type name_one: str
        :param name_two: name of node
        :type name_two: str
        """
        lca_coord = self.right_intersect(*(self[name].coordinate for name in names))
        return self.coord(lca_coord).name

    def reduce(self, nodes):
        """reduce Reduce tree until all that remains are these nodes.
        
        :param nodes: Names of nodes to keep
        :type nodes: Iterable[str]
        """
        _requested_node_set = {self._format_node_name(node) for node in nodes}
        child_to_idx = {
            child: self._pointers[self._format_node_name(child)]
            for child in {
                _child
                for _node in self.pool[1:]
                for _child in _node.children
            }
        }
        def pop_nodes(*childs):
            nonlocal child_to_idx
            _objs = {}
            for child in childs:
                idx = child_to_idx[child]
                _objs[child] = self.pool.pop(idx)
                self.pool.insert(idx, None)
            return _objs

        node_to_parent = {}
        for node in self.pool[1:]:
            node_name = self._format_node_name(node.name)
            for child in node.children:
                if child not in _requested_node_set:
                    node_to_parent[child] = node_name
        def replace_parents_child(node, child):
            nonlocal node_to_parent
            try:
                parent_children_list = self[node_to_parent[node]].children
                parent_children_list[parent_children_list.index(node)] = child
            except KeyError:    # Root.
                pass

        # Reset node pool to reflect changes in tree structure.
        branch_indices = [i for i, node in enumerate(self.pool) if node.type == "Branch"]
        for branch_idx in branch_indices[::-1]:
            branch = self.pool[branch_idx]
            branch_name = self._format_node_name(branch.name)
            children = branch.children
            children_with_counts = [child for child in children if child in _requested_node_set]

            if not children_with_counts:
                # Make this branch into a leaf node.
                branch.type = "Leaf"
                pop_nodes(*children)
                branch.children.clear()
            elif len(children_with_counts) == 2:
                # Pass validity of this node - dichotomous tree structure depends on
                # this node being present.
                _requested_node_set.add(branch_name)
            elif branch_name not in _requested_node_set:
                # Only one child with counts is implied by the previous two cases.
                # Replace this node with the valid child node.
                valid_child_name, branch_dist = children_with_counts[0], branch.distance
                # Remove children from active node pool.
                removed_children = pop_nodes(*children)
                # Remove branch from node pool.
                self.pool.pop(branch_idx)
                # Insert replacement.
                removed_children[valid_child_name].distance += branch_dist
                self.pool.insert(branch_idx, removed_children[valid_child_name])
                child_to_idx[valid_child_name] = branch_idx
                # Set this node as replacement child of parent.
                replace_parents_child(branch_name, valid_child_name)
                
        # Reset data structures with the new index.
        self.pool = [node for node in self.pool if node is not None]
        self.update_via_pool(keep_names=True, map_children_objects=False)

    @staticmethod
    def right_intersect(*lists) -> list:
        lists = [l for l in lists if len(l) > 0]
        L = min(len(l) for l in lists)
        for i in range(L):
            rev_index = i + 1
            ref = lists[0][-rev_index]
            for other in lists[1:]:
                if other[-rev_index] != ref:
                    if i == 0:
                        return []
                    else:
                        return lists[0][-i:]
        else:
            return lists[0][-L:]

    def draw_results(self, file_path, out_dir, skiprows=None, groups=None, cutoff=None, cpm=None, colour_list=None,
                     name_to_taxon=None, use_phyla=False, keep_zeros=True, use_node_names=True, log_scores=False,
                     itol_mode=False, flat_colour=False, sep=","):
        counts = pd.read_csv(file_path, sep=sep, index_col=0, header=0, skiprows=skiprows)

        # Remove any columns that contain only zeros.
        counts = counts.loc[:, (counts > 0).any().values]
        # Modify the index with formatted names.
        counts.index = [self._format_node_name(v) for v in counts.index]

        if counts.shape[1]:
            self.draw_tree(out_dir, counts=counts, groups=groups, cutoff=cutoff, cpm=cpm, colour_list=colour_list,
                        name_to_taxon=name_to_taxon, use_phyla=use_phyla, keep_zeros=keep_zeros,
                        use_node_names=use_node_names, log_scores=log_scores, itol_mode=itol_mode, flat_colour=flat_colour)
        else:
            print("Skipping plotting of %s - no samples with counts in this matrix." % file_path)

    def draw_tree(self, out_dir, counts: pd.DataFrame, groups=None, cutoff=None, cpm=None, colour_list=None, name_to_taxon=None,
                  use_phyla=False, keep_zeros=True, use_node_names=True, log_scores=True, itol_mode=False, flat_colour=False):
        from expam.sequences import format_name

        """
        Take care of groupings.
        """
        if groups is None:
            groups = [(None, (section,)) for section in counts.columns]
        else:
            # Format group names.
            groups = [
                (
                    col,
                    tuple(
                        format_name(group_member, remove_comp=True)
                        for group_member in group
                    )
                )
                for col, group in groups
            ]

        # Combine counts within the group.
        for _, group in groups:
            if len(group) > 1:
                group_name = group[0]

                counts.loc[:, group_name] = counts[list(group)].sum(axis=1)
                counts.drop(labels=list(group[1:]), axis=1, inplace=True)

        # Remove any groups that weren't specified.
        all_groups = counts.columns.tolist()
        specified_groups = set(group[0] for _, group in groups)
        unspecified_groups = [group for group in all_groups if group not in specified_groups]

        counts.drop(labels=unspecified_groups, axis=1, inplace=True)

        """
        Employ cutoff.
        """
        if cutoff is None:
            if cpm is None:
                cpm = 0.0
            # Use cpm cutoff.
            min_ser = counts.sum(axis=0) * cpm / 1e6
        else:
            if cpm is not None:
                min_ser = counts.sum(axis=0) * cpm / 1e6
                min_ser.clip(lower=cutoff)
            else:
                # Use raw cutoff.
                min_ser = cutoff

        nodes_with_counts = counts.index[(counts >= min_ser).any(axis=1)].tolist()

        """
        Attempt pruning of tree.
        """
        self.reduce(nodes_with_counts)

        """
        Colour generation.
        """
        max_vector = tuple(counts.max())
        min_vector = tuple(int(v) for v in counts[counts > 0].min())
        colours = [colour for colour, _ in groups]
        colour_generator = ColourGenerator(
            colours,
            max_vector,
            min_vector,
            colour_list=colour_list,
            log_scores=log_scores
        )
 
        if itol_mode:
            itol_data = [
                "TREE_COLORS",
                "# Drag me onto the iTOL tree (in browser) to apply me!",
                "SEPARATOR TAB",
                "DATA",
            ]

            for node in counts.index:
                node_counts = counts.loc[node, :]

                if node_counts.any():
                    colour = colour_generator.generate(tuple(node_counts), flat_colour=flat_colour)
                    itol_data.append("%s\trange\t%s\t%s" % (node, colour, node.upper()))

            # Write itol_data and Newick tree to file.
            name_modifier = re.findall(r"phylotree_(\S+).pdf", os.path.basename(out_dir))[0]
            itol_dir = os.path.join(os.path.dirname(out_dir), 'itol_%s' % name_modifier)
            if not os.path.exists(itol_dir):
                os.mkdir(itol_dir)
                print("iTOL output directory created at %s." % itol_dir)

            # Write style file.
            itol_style_dir = os.path.join(itol_dir, 'style.txt')
            with open(itol_style_dir, 'w') as f:
                f.write('\n'.join(itol_data))

            print("iTOL style written to %s." % itol_style_dir)

            # Write Newick file.
            itol_tree_dir = os.path.join(itol_dir, 'itol_tree.nwk')
            with open(itol_tree_dir, 'w') as f:
                f.write(self.to_newick())

            print("iTOL tree written to %s. Put this in itol (in your browser)." % itol_tree_dir)

        else:
            # Draw with ete3.
            try:
                from ete3 import AttrFace, faces, Tree, TreeStyle, NodeStyle, TextFace
            except ModuleNotFoundError as e:
                print("The ETE3 package is not installed. Either install this module, or use the\n\t--itol flag if you wish to use our native iTOL integration.")
                print("Skipping plotting of %s..." % out_dir)
                return
            except ImportError as e:
                print("Could not import drawing attributes from ete3. Error raised:")
                print(traceback.format_exc())
                print("See FAQ section on GitHub for fix to this problem.\n\t(https://github.com/seansolari/expam#problems-during-installation)")
                print("Skipping plotting of %s..." % out_dir)
                return

            """
            Create PhyloTree.
            """
            newick_string = self.to_newick()
            tree = Tree(newick_string, format=1)

            """
            Function to render any given node.
            """

            def layout(node):
                nonlocal name_to_taxon

                if node.is_leaf():
                    if use_node_names:
                        faces.add_face_to_node(AttrFace('name', fsize=20), node, column=0, position='aligned')

                    if use_phyla:
                        for phyla, colour in PHYLA_COLOURS:
                            lineage = name_to_taxon[node.name]

                            if phyla in lineage:
                                tax_face = TextFace("   ")
                                tax_face.background.color = colour
                                faces.add_face_to_node(tax_face, node, column=1, position='aligned')

                if node.name in counts.index:
                    node_counts = counts.loc[node.name, :]

                    if node_counts.any():
                        colour = colour_generator.generate(tuple(node_counts), flat_colour=flat_colour)

                        ns = NodeStyle()
                        ns['bgcolor'] = colour
                        node.set_style(ns)

            """
            Render tree.
            """
            ts = TreeStyle()
            ts.mode = "c"
            ts.show_leaf_name = False
            ts.layout_fn = layout
            ts.force_topology = False
            ts.allow_face_overlap = False
            ts.draw_guiding_lines = True
            ts.root_opening_factor = 1
            tree.render(
                out_dir,
                tree_style=ts,
                dpi=300
            )
            print("Phylogenetic tree written to %s!" % out_dir)


class HexColour:
    def __init__(self, hex_string):
        self.colour = hex_string.lstrip("#")

    def __add__(self, other):
        rgb_self = self.to_rgb()
        rgb_other = other.to_rgb()

        rgb_mix = rgb_self + rgb_other
        return rgb_mix.to_hex()

    def __iadd__(self, other):
        mix = self + other
        self.colour = mix.colour

    def __str__(self):
        return "#%s" % self.colour

    def to_rgb(self):
        hx = self.colour
        assert len(hx) == 6

        rgb = tuple(int(hx[2 * i:2 * (i + 1)], 16) for i in range(3))
        return RGBColour(rgb)

    def opaque(self, q):
        if q == 1:
            return self

        rgb = self.to_rgb().opaque(q)

        return rgb.to_hex()


class RGBColour:
    def __init__(self, tup):
        self.colour = tup

    def __add__(self, other):
        assert len(self.colour) == len(other.colour) == 3

        rgb = tuple(
            floor(0.5 * (self.colour[i] + other.colour[i]))
            for i in range(3)
        )
        return RGBColour(rgb)

    def __iadd__(self, other):
        mix = self + other
        self.colour = mix.colour

    @staticmethod
    def _format_rgb(num):
        f_hx = hex(num).lstrip('0x').upper()

        if len(f_hx) < 2:
            f_hx = ('0' * (2 - len(f_hx))) + f_hx

        return f_hx

    def to_hex(self):
        rgb = self.colour
        assert len(rgb) == 3

        hx = "#" + "".join(self._format_rgb(factor) for factor in rgb)
        return HexColour(hx)

    def opaque(self, q):
        return RGBColour(tuple(
            floor(255 - (q * (255 - v)))
            for v in self.colour
        ))


class RandomColour:
    def __init__(self, rng=None, seed=1):
        self.alpha = "0123456789ABCDEF"

        if rng is not None:
            if not isinstance(rng, random.Random):
                raise TypeError("rng must be instance of random.Random!")

            self.rng = rng
        else:
            if seed is not None:
                self.rng = random.Random(seed)
            else:
                self.rng = random.Random()

    def __next__(self):
        return "#" + "".join(self.rng.choices(self.alpha, k=6))


class ColourGenerator:
    def __init__(self, colours, max_vector, min_vector=None, colour_list=None, log_scores=False):
        """

        :param colours:
        :param max_vector:
        :param colour_list:
        """
        # Ensure we have enough colours for plotting.
        g = RandomColour()

        # Ensure we have enough colours for each group.
        required_colours = sum(colour is None for colour in colours)
        colours = [colour for colour in colours if colour is not None]

        if required_colours > 0:
            if colour_list is None:
                colour_list = [next(g) for _ in range(required_colours)]

            else:
                colours_to_generate = required_colours - len(colour_list)
                colour_list.extend(
                    [next(g) for _ in range(colours_to_generate)]
                )

            colours.extend(colour_list)

        self.colours = [HexColour(colour) for colour in colours]

        #
        self.max_vector = max_vector
        self.min_vector = min_vector
        self.make_log_score = log_scores

        if self.make_log_score and self.min_vector is None:
            raise AttributeError("Log scores requested but not min vector supplied!")

        # Calculate normalisation factors.
        self._log_norm_factor = [log(mx / mn) for mx, mn in zip(self.max_vector, self.min_vector)] if self.make_log_score else None

    def generate(self, score_vector, flat_colour = False):
        colour = HexColour("#FFFFFF")  # White background.
        scorer = self._linear_scores if not self.make_log_score else self._log_scores

        if not flat_colour:
            for next_colour, opacity in scorer(score_vector):
                colour = colour + next_colour.opaque(opacity)
        else:
            for next_colour, _ in scorer(score_vector):
                colour = colour + next_colour

        return str(colour)

    def _linear_scores(self, vector):
        for i, score in enumerate(vector):
            if score > 0:
                yield self.colours[i], score / self.max_vector[i]

    def _log_scores(self, vector):
        for i, score in enumerate(vector):
            if score > 0:
                yield self.colours[i], log(score / self.min_vector[i]) / self._log_norm_factor[i]



from collections import namedtuple
import logging
import os
import numpy as np


KEYS_DATA_TYPE = np.uint64
KEYS_DATA_TYPE_SIZE = np.dtype(KEYS_DATA_TYPE).itemsize
KEYS_DTYPE_STR = "uint64"

MAP_DATA_TYPE = np.uint16
MAP_DATA_TYPE_SIZE = np.dtype(MAP_DATA_TYPE).itemsize
MAP_DTYPE_STR = "uint16"

DataTypeConfig = namedtuple('DataTypeConfig', ['keys_dtype', 'keys_dtype_size', 'keys_dtype_str', 'map_dtype', 'map_dtype_size', 'map_dtype_str'])
expam_dtypes = DataTypeConfig(KEYS_DATA_TYPE, KEYS_DATA_TYPE_SIZE, KEYS_DTYPE_STR, MAP_DATA_TYPE, MAP_DATA_TYPE_SIZE, MAP_DTYPE_STR)

DATABASE_RELATIVE_PATH = "database"
PHYLOGENY_RELATIVE_PATH = "phylogeny"
LOG_RELATIVE_PATH = "logs"
CONF_RELATIVE_PATH = "conf.json"

ACCESSION_ID_RELATIVE_PATH = os.path.join(PHYLOGENY_RELATIVE_PATH, "accession_ids.csv")
TAXID_LINEAGE_MAP_RELATIVE_PATH = os.path.join(PHYLOGENY_RELATIVE_PATH, "taxid_lineage.csv")
TAXON_RANK_MAP_RELATIVE_PATH = os.path.join(PHYLOGENY_RELATIVE_PATH, "taxa_rank.csv")

LCA_MATRIX_RELATIVE_PATH = os.path.join(PHYLOGENY_RELATIVE_PATH, "lca_matrix.npy")
DATABASE_FILE_RELATIVE_PATH = os.path.join(DATABASE_RELATIVE_PATH, "expam_db.h5")

FileLocationConfig = namedtuple(
    'FileLocationConfig',
    [
        'base', 'database', 'phylogeny', 'logs', 'conf',
        'accession_id', 'taxid_lineage', 'taxon_rank',
        'lca_matrix',
        'database_file'
    ]
)

CHUNK_SIZE = 32
UNION_RATIO = 4 / 5

TIMEOUT = 1e-4

NULL_VALUE = 0  # Return value for foreign kmer.
EXPAM_TEMP_EXT = "expam_temp"

LOG_FORMAT = logging.Formatter(
    fmt='%(asctime)s... %(message)s',
    datefmt='%m/%d/%Y %I:%M:%S %p'
)




from collections import namedtuple

import os


ENTREZ_EMAIL = "sean.solari@monash.edu"
ENTREZ_TOOL = "expam"
ENTREZ_API = "8e5edfa1413576bca4f48b4a5520f6295308"
_STEP = 500  # NCBI Database searches at a time.

PHY_RESULTS = "phy"
TAX_RESULTS = "tax"
RAW_RESULTS = "raw"
TEMP_RESULTS = "temp"

CLASSIFIED_NAME = "classified.csv"
SPLIT_NAME = "split.csv"

PHY_RAW = os.path.join(PHY_RESULTS, RAW_RESULTS)
TAX_RAW = os.path.join(TAX_RESULTS, RAW_RESULTS)
PHY_CLASSIFIED_FILE = os.path.join(PHY_RESULTS, CLASSIFIED_NAME)
PHY_SPLIT_FILE = os.path.join(PHY_RESULTS, SPLIT_NAME)
TAX_CLASSIFIED_FILE = os.path.join(TAX_RESULTS, CLASSIFIED_NAME)
TAX_SPLIT_FILE = os.path.join(TAX_RESULTS, SPLIT_NAME)


ResultsPathConfig = namedtuple(
    'ResultsPathConfig',
    [
        'base', 'phy', 'tax',
        'temp',
        'phy_raw', 'tax_raw',
        'phy_classified', 'phy_split',
        'tax_classified', 'tax_split'
    ]
)
